# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Functions for processing acoustic data """

import os
import numpy as np
import h5py as h5
import libtfr


os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"

__version__ = "20220808-1"

